import base64
from enum import Enum
from typing import Optional

import httpx
import requests
from nucliadb_models.search import ChatRequest
from nucliadb_sdk import NucliaDB, Region

from nuclia.lib.utils import handle_http_errors

RESOURCE_PATH = "/resource/{rid}"
RESOURCE_PATH_BY_SLUG = "/slug/{slug}"
SEARCH_PATH = "/search"
CREATE_RESOURCE_PATH = "/resources"
CREATE_VECTORSET = "/vectorset/{vectorset}"
VECTORSETS = "/vectorsets"
COUNTER = "/counters"
SEARCH_URL = "/search"
FIND_URL = "/find"
CHAT_URL = "/chat"
LABELS_URL = "/labelsets"
ENTITIES_URL = "/entitiesgroups"
DOWNLOAD_URL = "/{uri}"
TUS_UPLOAD_RESOURCE_URL = "/resource/{rid}/file/{field}/tusupload"
TUS_UPLOAD_URL = "/tusupload"


class Environment(str, Enum):
    CLOUD = "CLOUD"
    OSS = "OSS"


class NucliaDBClient:
    environment: Environment
    base_url: str
    api_key: Optional[str]
    url: Optional[str] = None
    reader_session: Optional[httpx.Client] = None
    stream_session: Optional[requests.Session] = None
    writer_session: Optional[httpx.Client] = None

    def __init__(
        self,
        *,
        environment: Environment = Environment.CLOUD,
        url: Optional[str] = None,
        api_key: Optional[str] = None,
        user_token: Optional[str] = None,
        region: Optional[str] = None,
        base_url: Optional[str] = None,
    ):
        if environment == Environment.OSS:
            region_obj = Region.ON_PREM
        else:
            region_obj = Region(region)

        headers = {}
        if user_token is not None:
            headers["Authorization"] = f"Bearer {user_token}"
        headers["X-SYNCHRONOUS"] = "True"
        if base_url is not None:
            v2url = base_url
        if base_url is None and url is not None:
            v2url = "/".join(url.split("/")[:-3])
        self.base_url = v2url
        self.ndb = NucliaDB(
            region=region_obj, url=v2url, api_key=api_key, headers=headers
        )
        self.api_key = api_key
        self.environment = environment

        if url is not None:
            self.kbid = url.strip("/").split("/")[-1]
            self.url = url

        if environment == Environment.CLOUD and api_key is not None:
            reader_headers = {
                "X-NUCLIA-SERVICEACCOUNT": f"Bearer {api_key}",
            }
            writer_headers = {
                "X-NUCLIA-SERVICEACCOUNT": f"Bearer {api_key}",
                "X-SYNCHRONOUS": "True",
            }
        elif (
            environment == Environment.CLOUD
            and api_key is None
            and user_token is not None
        ):
            reader_headers = {
                "Authorization": f"Bearer {user_token}",
            }
            writer_headers = {
                "Authorization": f"Bearer {user_token}",
                "X-SYNCHRONOUS": "True",
            }
        elif (
            environment == Environment.CLOUD and api_key is None and user_token is None
        ):
            # Public
            reader_headers = {}
            writer_headers = {}
        else:
            reader_headers = {
                "X-NUCLIADB-ROLES": f"READER",
            }
            writer_headers = {
                "X-NUCLIADB-ROLES": f"WRITER",
                "X-SYNCHRONOUS": "True",
            }

        if url is not None:
            self.reader_session = httpx.Client(
                headers=reader_headers, base_url=url  # type: ignore
            )
            self.stream_session = requests.Session()
            self.stream_session.headers.update(reader_headers)
            self.writer_session = httpx.Client(
                headers=writer_headers, base_url=url  # type: ignore
            )

    def chat(self, request: ChatRequest):
        if self.url is None or self.stream_session is None:
            raise Exception("KB not configured")
        url = f"{self.url}{CHAT_URL}"
        response: requests.Response = self.stream_session.post(
            url, data=request.json(), stream=True
        )
        handle_http_errors(response)
        return response

    def download(self, uri: str) -> bytes:
        # uri has format
        # /kb/2a00d5b4-cfcc-48eb-85ac-d70bfd38b26d/resource/41d02aac4ade48098b23e38141807738/file/file/download/field
        # we need to remove the kb url
        if self.reader_session is None:
            raise Exception("KB not configured")

        uri_parts = uri.split("/")
        if len(uri_parts) < 9:
            raise AttributeError("Not a valid download uri")

        new_uri = "/".join(uri_parts[3:])
        url = DOWNLOAD_URL.format(uri=new_uri)
        response: httpx.Response = self.reader_session.get(url)
        handle_http_errors(response)
        return response.content

    def start_tus_upload(
        self,
        size: int,
        filename: str,
        field: Optional[str] = None,
        rid: Optional[str] = None,
        md5: Optional[str] = None,
        content_type: str = "application/octet-stream",
    ):
        if self.writer_session is None:
            raise Exception("KB not configured")

        if rid is not None:
            if field is None:
                field = filename
            url = TUS_UPLOAD_RESOURCE_URL.format(rid=rid, field=field)
        else:
            url = TUS_UPLOAD_URL
        encoded_filename = base64.b64encode(filename.encode()).decode()
        headers = {
            "upload-length": str(size),
            "tus-resumable": "1.0.0",
            "upload-metadata": f"filename {encoded_filename}",
            "content-type": content_type,
        }
        if md5 is not None:
            headers[
                "upload-metadata"
            ] += f",md5 {base64.b64encode(md5.encode()).decode()}"

        response: httpx.Response = self.writer_session.post(url, headers=headers)
        handle_http_errors(response)
        return response.headers.get("Location")

    def patch_tus_upload(self, upload_url: str, data: bytes, offset: int) -> int:
        if self.writer_session is None:
            raise Exception("KB not configured")

        headers = {
            "upload-offset": str(offset),
        }
        # upload url has all path, we should remove /kb/kbid/
        upload_url = "/" + "/".join(upload_url.split("/")[3:])
        response: httpx.Response = self.writer_session.patch(
            upload_url, headers=headers, content=data
        )
        handle_http_errors(response)
        return int(response.headers.get("Upload-Offset"))



import base64
from time import sleep
from typing import Dict, List, Optional

import requests
from nucliadb_protos.writer_pb2 import BrokerMessage
from nuclia.sdk.logger import logger

from nuclia import REGIONAL
from nuclia.exceptions import NuaAPIException
from nuclia.lib.nua_responses import (
    Author,
    ChatModel,
    LearningConfig,
    Message,
    ProcessingStatus,
    PublicPushPayload,
    PublicPushResponse,
    PullResponse,
    PullStatus,
    Sentence,
    Source,
    SummarizedModel,
    SummarizeModel,
    SummarizeResource,
    Tokens,
    UserPrompt,
)

SENTENCE_PREDICT = "/api/v1/predict/sentence"
CHAT_PREDICT = "/api/v1/predict/chat"
SUMMARIZE_PREDICT = "/api/v1/predict/summarize"
TOKENS_PREDICT = "/api/v1/predict/tokens"
PUSH_PROCESS = "/api/v1/processing/push"
PULL_PROCESS = "/api/v1/processing/pull"
UPLOAD_PROCESS = "/api/v1/processing/upload"
STATUS_PROCESS = "/api/v1/processing/status"


class NuaClient:
    def __init__(self, region: str, account: str, token: str):
        self.region = region
        self.account = account
        self.token = token
        if "http" in region:
            self.url = region.strip("/")
        else:
            self.url = REGIONAL.format(region=region).strip("/")
        self.headers = {"X-STF-NUAKEY": f"Bearer {token}"}

    def sentence_predict(self, text: str, model: Optional[str] = None) -> Sentence:
        endpoint = f"{self.url}{SENTENCE_PREDICT}?text={text}"
        if model:
            endpoint += f"&model={model}"
        resp = requests.get(endpoint, headers=self.headers)
        if resp.status_code == 200:
            return Sentence.parse_obj(resp.json())
        else:
            import pdb

            pdb.set_trace()
            raise NuaAPIException()

    def tokens_predict(self, text: str, model: Optional[str] = None) -> Tokens:
        endpoint = f"{self.url}{TOKENS_PREDICT}?text={text}"
        if model:
            endpoint += f"&model={model}"
        resp = requests.get(endpoint, headers=self.headers)
        if resp.status_code == 200:
            return Tokens.parse_obj(resp.json())
        else:
            raise NuaAPIException()

    def generate_predict(self, text: str, model: Optional[str] = None) -> bytes:
        endpoint = f"{self.url}{CHAT_PREDICT}"
        if model:
            endpoint += f"?model={model}"

        body = ChatModel(
            question="",
            retrieval=False,
            user_id="Nuclia PY CLI",
            user_prompt=UserPrompt(prompt=text),
        )
        resp = requests.post(
            endpoint, data=body.json(), headers=self.headers, stream=True
        )
        if resp.status_code == 200:
            response = b""
            for chunk in resp.raw.stream(1000, decode_content=True):
                response += chunk

            return response
        else:
            raise NuaAPIException()

    def summarize(
        self, documents: Dict[str, str], model: Optional[str] = None
    ) -> SummarizedModel:
        endpoint = f"{self.url}{SUMMARIZE_PREDICT}"
        if model:
            endpoint += f"?model={model}"

        body = SummarizeModel(
            resources={
                key: SummarizeResource(fields={"field": document})
                for key, document in documents.items()
            }
        )
        resp = requests.post(
            endpoint, data=body.json(), headers=self.headers, stream=True
        )

        if resp.status_code == 200:
            return SummarizedModel.parse_raw(resp.content)
        else:
            raise NuaAPIException()

    def generate_retrieval(
        self, question: str, context: List[str], model: Optional[str] = None
    ) -> bytes:
        endpoint = f"{self.url}{CHAT_PREDICT}"
        if model:
            endpoint += f"?model={model}"

        message_context = [
            Message(author=Author.USER, text=message) for message in context
        ]
        body = ChatModel(
            question=question,
            retrieval=True,
            user_id="Nuclia PY CLI",
            context=message_context,
        )
        resp = requests.post(
            endpoint, data=body.json(), headers=self.headers, stream=True
        )

        if resp.status_code == 200:
            response = b""
            for chunk in resp.raw.stream(1000, decode_content=True):
                response += chunk

            return response
        else:
            raise NuaAPIException()

    def process_file(
        self, path: str, config: Optional[LearningConfig] = None
    ) -> PublicPushResponse:
        filename = path.split("/")[-1]
        upload_endpoint = f"{self.url}{UPLOAD_PROCESS}"

        headers = self.headers.copy()
        headers["X-FILENAME"] = base64.b64encode(filename.encode()).decode()
        with open(path, "rb") as file_to_upload:
            data = file_to_upload.read()

        resp = requests.post(upload_endpoint, data=data, headers=headers)

        # file_token =
        payload = PublicPushPayload(
            uuid=None, source=Source.HTTP, learning_config=config
        )
        payload.filefield[filename] = resp.content.decode()
        process_endpoint = f"{self.url}{PUSH_PROCESS}"
        resp = requests.post(
            process_endpoint, data=payload.json(), headers=self.headers
        )
        if resp.status_code == 200:
            return PublicPushResponse.parse_raw(resp.content)
        else:
            raise NuaAPIException()

    def wait_for_processing(
        self, response: Optional[PublicPushResponse] = None, timeout: int = 10
    ) -> Optional[BrokerMessage]:
        collect_endpoint = f"{self.url}{PULL_PROCESS}"

        pull_resp = PullResponse(status=PullStatus.EMPTY, payload=None, msgid=None)
        count = 0
        bm = None
        found = False
        while found is False and count < timeout * 20:
            resp = requests.get(collect_endpoint, headers=self.headers)
            pull_resp = PullResponse.parse_raw(resp.content)
            count += 1
            if pull_resp.status == PullStatus.EMPTY:
                sleep(3)
            elif pull_resp.payload is not None:
                bm = BrokerMessage()
                bm.ParseFromString(base64.b64decode(pull_resp.payload))
                if response is not None and response.account_seq is not None:
                    if bm.account_seq == response.account_seq:
                        found = True
                    else:
                        logger.info(
                            f"Received account seq {bm.account_seq} but waiting for {response.account_seq}, "
                            "old processing info or a shared NUA key"
                        )
        return bm

    def processing_status(self) -> ProcessingStatus:
        activity_endpoint = f"{self.url}{STATUS_PROCESS}"
        resp = requests.get(activity_endpoint, headers=self.headers)
        return ProcessingStatus.parse_raw(resp.content)



from nuclia.sdk.predict import NucliaPredict


def test_predict(testing_config):
    np = NucliaPredict()
    embed = np.sentence(text="This is my text", model="multilingual-2023-02-21")
    assert embed.time > 0
    assert len(embed.data) == 768



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from nuclia.config import Config
    from nuclia.sdk.auth import NucliaAuth


@dataclass
class DataConfig:
    config: Optional[Config] = None
    auth: Optional[NucliaAuth] = None


DATA = DataConfig()


def set_config(config: Config):
    DATA.config = config


def get_config() -> Config:
    if DATA.config is None:
        from nuclia.config import read_config

        DATA.config = read_config()
    return DATA.config


def get_auth() -> NucliaAuth:
    get_config()
    if DATA.auth is None:
        from nuclia.sdk.auth import NucliaAuth

        DATA.auth = NucliaAuth()
    return DATA.auth



from typing import Optional

from nuclia.data import get_auth
from nuclia.decorators import nua
from nuclia.lib.nua import NuaClient
from nuclia.lib.nua_responses import LearningConfig, ProcessingStatus
from nuclia.sdk.auth import NucliaAuth
from nucliadb_protos.writer_pb2 import BrokerMessage


class NucliaProcessing:
    @property
    def _auth(self) -> NucliaAuth:
        auth = get_auth()
        return auth

    @nua
    def process_file(
        self, path: str, config: Optional[LearningConfig] = None, **kwargs
    ) -> Optional[BrokerMessage]:
        nc: NuaClient = kwargs["nc"]
        response = nc.process_file(path, config)
        payload = nc.wait_for_processing(response)
        return payload

    @nua
    def status(self, **kwargs) -> ProcessingStatus:
        nc: NuaClient = kwargs["nc"]
        return nc.processing_status()



import os
from typing import List, Optional

BASE_DOMAIN = os.environ.get("BASE_NUCLIA_DOMAIN", "nuclia.cloud")
BASE = f"https://{BASE_DOMAIN}"
REGIONAL = "https://{region}." + BASE_DOMAIN
CLOUD_ID = BASE.split("/")[-1]

USE_NEW_REGIONAL_ENDPOINTS = os.environ.get("USE_NEW_REGIONAL_ENDPOINTS", "") == "TRUE"


def get_global_url(path: str):
    return BASE + path


def get_regional_url(region: str, path: str):
    return REGIONAL.format(region=region) + path


def get_list_parameter(param: Optional[List[str]]) -> List[str]:
    param = list(param or [])
    if len(param) > 0 and all([len(el) == 1 for el in param]):
        # Python Fire converts single string to list of chars
        param = ["".join(param)]
    return param



import re
from datetime import datetime
from enum import Enum
from typing import Dict, List, Optional, cast

from pydantic import BaseModel, ConstrainedStr, Field


class Sentence(BaseModel):
    data: List[float]
    time: float


class Author(str, Enum):
    NUCLIA = "NUCLIA"
    USER = "USER"


class Message(BaseModel):
    author: Author
    text: str


class UserPrompt(BaseModel):
    prompt: str


class ChatModel(BaseModel):
    question: str
    retrieval: bool = True
    user_id: str
    system: Optional[str] = None
    chat_history: List[Message] = []
    context: List[Message] = []
    query_context: List[str] = []
    truncate: Optional[bool] = False
    user_prompt: Optional[UserPrompt] = None


class Token(BaseModel):
    text: str
    ner: str
    start: int
    end: int


class Tokens(BaseModel):
    tokens: List[Token]
    time: float


class SummarizeResource(BaseModel):
    fields: Dict[str, str]


class SummarizeModel(BaseModel):
    resources: Dict[str, SummarizeResource]


class SummarizedResource(BaseModel):
    summary: str
    tokens: int


class SummarizedModel(BaseModel):
    resources: Dict[str, SummarizedResource]
    summary: str = ""


class RephraseModel(BaseModel):
    question: str
    chat_history: List[Message] = []
    context: List[Message] = []
    user_id: str


class WebhookConfig(BaseModel):
    uri: Optional[str] = None
    headers: Dict[str, str] = {}


# Pull models


class PullStatus(str, Enum):
    OK = "ok"
    ERROR = "error"
    EMPTY = "empty"


class PullResponse(BaseModel):
    status: PullStatus
    payload: Optional[bytes]
    msgid: Optional[str]


# Push models


class ProcessingQueueType(str, Enum):
    SHARED = "shared"
    PRIVATE = "private"


class Format(int, Enum):
    PLAIN = 0
    HTML = 1
    MARKDOWN = 2
    RST = 3
    JSON = 4


class ProcessingMessageContent(BaseModel):
    text: Optional[str] = None
    format: Format
    attachments: Optional[List[str]] = None


class ProcessingMessage(BaseModel):
    timestamp: Optional[datetime] = None
    who: Optional[str] = None
    to: List[str] = []
    content: ProcessingMessageContent
    ident: str


class ProcessingConversation(BaseModel):
    messages: List[ProcessingMessage]


class LayoutFormat(int, Enum):
    FLAPSv1 = 0


class TypeBlock(int, Enum):
    TITLE = 0
    DESCRIPTION = 1
    RICHTEXT = 2
    TEXT = 3
    ATTACHMENTS = 4
    COMMENTS = 5
    CLASSIFICATIONS = 6


class ProcessingBlock(BaseModel):
    x: int
    y: int
    cols: int
    rows: int
    type: TypeBlock
    ident: str
    payload: str
    file: Optional[str] = None


class ProcessingLayoutDiff(BaseModel):
    format: LayoutFormat
    blocks: Dict[str, ProcessingBlock]


class LinkUpload(BaseModel):
    link: str
    headers: Dict[str, str]
    cookies: Dict[str, str]
    localstorage: Dict[str, str]


class Text(BaseModel):
    body: str
    format: Format


class Source(int, Enum):
    HTTP = 0
    INGEST = 1


class RestrictedIDString(ConstrainedStr):
    regex = re.compile(r"^[a-z0-9_-]+$")
    min_length = 1


class PushProcessingOptions(BaseModel):
    # Enable ML processing
    ml_text: bool = True


class LearningConfig(BaseModel):
    semantic_model: Optional[str] = None
    generative_model: Optional[str] = None
    ner_model: Optional[str] = None
    anonymization_model: Optional[str] = None
    visual_labeling: Optional[str] = None


class PushPayload(BaseModel):
    kbid: Optional[RestrictedIDString] = cast(RestrictedIDString, "default")
    uuid: Optional[RestrictedIDString] = None

    slug: Optional[str] = None
    source: Optional[Source] = None
    userid: Optional[str] = None

    # Generic Field
    genericfield: Dict[str, Text] = {}

    # New File
    filefield: Dict[str, str] = {}

    # New Link
    linkfield: Dict[str, LinkUpload] = {}

    # Diff on Text Field
    textfield: Dict[str, Text] = {}

    # Diff on a Layout Field
    layoutfield: Dict[str, ProcessingLayoutDiff] = {}

    # New conversations to process
    conversationfield: Dict[str, ProcessingConversation] = {}

    # List of available processing options (with default values)
    processing_options: Optional[PushProcessingOptions] = Field(
        default_factory=PushProcessingOptions
    )

    learning_config: Optional[LearningConfig] = None


class PublicPushPayload(PushPayload):
    # If provided, will override the webhookconfig on the nua-token (if any)
    webhook_config: Optional[WebhookConfig] = None


class PublicPushResponse(BaseModel):
    seqid: Optional[int] = None
    account_seq: Optional[int] = None
    queue: Optional[ProcessingQueueType] = None
    # On Public api, uuid may be generated by the proxy, so we need to return it to the user
    uuid: str


class ProcessingStatusInfo(BaseModel):
    last_delivered_seqid: Optional[
        int
    ] = None  # When none, means we already don't have information about this queue


class ProcessingStatus(BaseModel):
    shared: ProcessingStatusInfo
    account: Optional[ProcessingStatusInfo]



from uuid import uuid4

import pytest

from nuclia.sdk.kbs import NucliaKBS
from nuclia.tests.fixtures import IS_PROD, TESTING_ACCOUNT_SLUG, TESTING_KBID

NEW_KB_SLUG = "testkb-" + uuid4().hex


def test_list_kbs(testing_config):
    kbs = NucliaKBS()
    all = kbs.list()
    assert TESTING_KBID in [kb.id for kb in all]


def test_add_and_delete_kb(testing_config):
    if IS_PROD:
        # not possible on our prod account
        assert True
        return
    kbs = NucliaKBS()
    kb = kbs.add(account=TESTING_ACCOUNT_SLUG, slug=NEW_KB_SLUG, title="Test KB", zone="europe-1")
    assert kb["id"] is not None
    assert kb["slug"] == NEW_KB_SLUG
    assert kb["title"] == "Test KB"
    assert kbs.get(account=TESTING_ACCOUNT_SLUG, slug=NEW_KB_SLUG, zone="europe-1") is not None

    kbs.delete(account=TESTING_ACCOUNT_SLUG, id=kb["id"], zone="europe-1")
    with pytest.raises(ValueError):
        kbs.get(account=TESTING_ACCOUNT_SLUG, slug=NEW_KB_SLUG, zone="europe-1")



from asyncio import BaseEventLoop
from functools import partial

import pytest
from nucliadb_models.resource import KnowledgeBoxList, KnowledgeBoxObj, ResourceList
from nucliadb_sdk.tests.fixtures import NucliaFixture

from nuclia.sdk import NucliaAuth, NucliaDB, NucliaKB


@pytest.mark.asyncio
async def test_crud(nucliadb: NucliaFixture, event_loop: BaseEventLoop):
    auth = NucliaAuth()
    nkb = NucliaKB()
    ndb = NucliaDB()

    await event_loop.run_in_executor(
        None, auth.nucliadb, f"http://localhost:{nucliadb.port}"
    )

    listing_kbs: KnowledgeBoxList = await event_loop.run_in_executor(None, ndb.list)
    assert len(listing_kbs.kbs) == 0
    creation = partial(
        ndb.create, slug="kb", title="title", similarity="dot", default=True
    )

    kb_obj: KnowledgeBoxObj = await event_loop.run_in_executor(None, creation)
    assert kb_obj.slug == "kb"

    listing_kbs = await event_loop.run_in_executor(None, ndb.list)
    assert len(listing_kbs.kbs) == 1

    listing = partial(nkb.list, interactive=False)
    listing_res: ResourceList = await event_loop.run_in_executor(None, listing)
    assert len(listing_res.resources) == 0



class ConfigNotFound(Exception):
    pass


class NeedUserToken(Exception):
    pass


class UserTokenExpired(Exception):
    pass


class NotDefinedDefault(Exception):
    pass


class NuaAPIException(Exception):
    pass


class AlreadyConsumed(Exception):
    pass
